@@ -32,7 +32,8 @@
 #include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
 #include <pango/pango-tabs.h>
-#include <glib/gi18n.h>
+
+#include "gtksourceview-i18n.h"
 
 #include "gtksourceview-marshal.h"
 #include "gtksourceview.h"
@@ -42,12 +43,22 @@
 */
 #undef ENABLE_DEBUG
 
+/*
+#define ENABLE_PROFILE
+*/
+#undef ENABLE_PROFILE
+
 #ifdef ENABLE_DEBUG
 #define DEBUG(x) (x)
 #else
 #define DEBUG(x)
 #endif
 
+#ifdef ENABLE_PROFILE
+#define PROFILE(x) (x)
+#else
+#define PROFILE(x)
+#endif
 
 #define COMPOSITE_ALPHA                 225
 #define GUTTER_PIXMAP 			16
@@ -76,7 +87,8 @@ enum {
 	PROP_SHOW_MARGIN,
 	PROP_MARGIN,
 	PROP_SMART_HOME_END,
-	PROP_HIGHLIGHT_CURRENT_LINE
+	PROP_HIGHLIGHT_CURRENT_LINE,
+	PROP_INDENT_ON_TAB
 };
 
 struct _GtkSourceViewPrivate
@@ -88,6 +100,7 @@ struct _GtkSourceViewPrivate
 	gboolean	 insert_spaces;
 	gboolean	 show_margin;
 	gboolean	 highlight_current_line;
+	gboolean	 indent_on_tab;
 	guint		 margin;
 	gint             cached_margin_width;
 	gboolean	 smart_home_end;
@@ -103,12 +116,10 @@ typedef enum {
 	TARGET_COLOR = 200
 } GtkSourceViewDropTypes;
 
-static GtkTargetEntry drop_types[] = {
+static const GtkTargetEntry drop_types[] = {
 	{"application/x-color", 0, TARGET_COLOR}
 };
 
-static gint n_drop_types = sizeof (drop_types) / sizeof (drop_types[0]);
-
 static guint signals[LAST_SIGNAL] = { 0 };
 
 static GObjectClass *parent_class = NULL;
@@ -143,8 +154,10 @@ static void 	gtk_source_view_get_lines 		(GtkTextView       *text_view,
 				       			 gint              *countp);
 static gint     gtk_source_view_expose 			(GtkWidget         *widget,
 							 GdkEventExpose    *event);
-static gboolean	key_press_event				(GtkWidget         *widget, 
+static gboolean	gtk_source_view_key_press_event		(GtkWidget         *widget,
 							 GdkEventKey       *event);
+static gboolean	gtk_source_view_button_press_event	(GtkWidget         *widget,
+							 GdkEventButton    *event);
 static void 	view_dnd_drop 				(GtkTextView       *view, 
 							 GdkDragContext    *context,
 							 gint               x,
@@ -189,7 +202,8 @@ gtk_source_view_class_init (GtkSourceViewClass *klass)
 	object_class->get_property = gtk_source_view_get_property;
 	object_class->set_property = gtk_source_view_set_property;
 
-	widget_class->key_press_event = key_press_event;
+	widget_class->key_press_event = gtk_source_view_key_press_event;
+	widget_class->button_press_event = gtk_source_view_button_press_event;
 	widget_class->expose_event = gtk_source_view_expose;
 	widget_class->style_set = gtk_source_view_style_set;
 
@@ -276,6 +290,101 @@ gtk_source_view_class_init (GtkSourceViewClass *klass)
 							       FALSE,
 							       G_PARAM_READWRITE));
 
+	g_object_class_install_property (object_class,
+					 PROP_INDENT_ON_TAB,
+					 g_param_spec_boolean ("indent_on_tab",
+							       _("Indent on tab"),
+							       _("Whether to indent the selected text when the tab key is pressed"),
+							       FALSE,
+							       G_PARAM_READWRITE));
+
+	/**
+	 * GtkSourceView:right-margin-line-alpha:
+	 *
+	 * The ::right-margin-line-alpha determines the alpha component with
+	 * which the vertical line will be drawn. 0 means it is fully transparent
+	 * (invisible). 255 means it has full opacity (text under the line won't
+	 * be visible).
+	 *
+	 * Since: 1.6
+	 */
+	gtk_widget_class_install_style_property (widget_class,
+		g_param_spec_int ("right-margin-line-alpha",
+				  _("Margin Line Alpha"),
+				  _("Transparency of the margin line"),
+				  0,
+				  255,
+				  40,
+				  G_PARAM_READABLE));
+
+	/**
+	 * GtkSourceView:right-margin-line-color:
+	 *
+	 * The ::right-margin-line-color property contains the color with
+	 * which the right margin line (vertical line indicating the right
+	 * margin) will be drawn.
+	 *
+	 * Since: 1.6
+	 */
+	gtk_widget_class_install_style_property (widget_class,
+		g_param_spec_boxed ("right-margin-line-color",
+				    _("Margin Line Color"),
+				    _("Color to use for the right margin line"),
+				    GDK_TYPE_COLOR,
+				    G_PARAM_READABLE));
+
+	/**
+	 * GtkSourceView:right-margin-overlay-toggle:
+	 *
+	 * The ::right-margin-overlay-toggle property determines whether the
+	 * widget will draw a transparent overlay on top of the area on the
+	 * right side of the right margin line. On some systems, this has a
+	 * noticable performance impact, so this property is FALSE by default.
+	 *
+	 * Since: 1.6
+	 */
+	gtk_widget_class_install_style_property (widget_class,
+		g_param_spec_string ("right-margin-overlay-toggle",
+				     _("Margin Overlay Toggle"),
+				     _("Whether to draw the right margin overlay"),
+				     "FALSE",
+				     G_PARAM_READABLE));
+
+	/**
+	 * GtkSourceView:right-margin-overlay-alpha:
+	 *
+	 * The ::right-margin-overlay-alpha determines the alpha component with
+	 * which the overlay will be drawn. 0 means it is fully transparent
+	 * (invisible). 255 means it has full opacity (text under the overlay
+	 * won't be visible).
+	 *
+	 * Since: 1.6
+	 */
+	gtk_widget_class_install_style_property (widget_class,
+		g_param_spec_int ("right-margin-overlay-alpha",
+				  _("Margin Overlay Alpha"),
+				  _("Transparency of the margin overlay"),
+				  0,
+				  255,
+				  15,
+				  G_PARAM_READABLE));
+
+	/**
+	 * GtkSourceView:right-margin-overlay-color:
+	 *
+	 * The ::right-margin-overlay-color property contains the color with
+	 * which the right margin overlay will be drawn. Setting this property
+	 * will only have an effect if ::right-margin-overlay-toggle is TRUE.
+	 *
+	 * Since: 1.6
+	 */
+	gtk_widget_class_install_style_property (widget_class,
+		g_param_spec_boxed ("right-margin-overlay-color",
+				    _("Margin Overlay Color"),
+				    _("Color to use for drawing the margin overlay"),
+				    GDK_TYPE_COLOR,
+				    G_PARAM_READABLE));
+
 	signals [UNDO] =
 		g_signal_new ("undo",
 			      G_TYPE_FROM_CLASS (klass),
@@ -372,7 +481,10 @@ gtk_source_view_set_property (GObject      *object,
 			gtk_source_view_set_highlight_current_line (view,
 								    g_value_get_boolean (value));
 			break;
-
+		case PROP_INDENT_ON_TAB:
+			gtk_source_view_set_indent_on_tab (view,
+							   g_value_get_boolean (value));
+			break;
 		
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -442,6 +554,12 @@ gtk_source_view_get_property (GObject    *object,
 		case PROP_HIGHLIGHT_CURRENT_LINE:
 			g_value_set_boolean (value,
 					     gtk_source_view_get_highlight_current_line (view));
+			break;
+
+		case PROP_INDENT_ON_TAB:
+			g_value_set_boolean (value,
+					     gtk_source_view_get_indent_on_tab (view));
+			break;
 
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -471,8 +589,8 @@ gtk_source_view_init (GtkSourceView *view)
 	tl = gtk_drag_dest_get_target_list (GTK_WIDGET (view));
 	g_return_if_fail (tl != NULL);
 
-	gtk_target_list_add_table (tl, drop_types, n_drop_types);
-	
+	gtk_target_list_add_table (tl, drop_types, G_N_ELEMENTS (drop_types));
+
 	g_signal_connect (G_OBJECT (view), 
 			  "drag_data_received", 
 			  G_CALLBACK (view_dnd_drop), 
@@ -1283,16 +1401,39 @@ gtk_source_view_expose (GtkWidget      *widget,
 					    redraw_rect.width,
 					    height);
 		}
+		
+		/* Have GtkTextView draw the text first. */
+		if (GTK_WIDGET_CLASS (parent_class)->expose_event)
+			event_handled = 
+				(* GTK_WIDGET_CLASS (parent_class)->expose_event)
+				(widget, event);
 
+		/* Draw the right margin vertical line + overlay. */
 		if (view->priv->show_margin && 
 		    (event->window == gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_TEXT)))
 		{
 			GdkRectangle visible_rect;
 			GdkRectangle redraw_rect;
+			cairo_t *cr;
+			double x;
+			GdkColor *line_color;
+			gchar *toggle;
+			guchar alpha;
+
+#ifdef ENABLE_PROFILE
+			static GTimer *timer = NULL;
+#endif
 
 			if (view->priv->cached_margin_width < 0)
 				view->priv->cached_margin_width =
 					calculate_real_tab_width (view, view->priv->margin, '_');
+
+#ifdef ENABLE_PROFILE
+			if (timer == NULL)
+				timer = g_timer_new ();
+				
+			g_timer_start (timer);
+#endif 
 		
 			gtk_text_view_get_visible_rect (text_view, &visible_rect);
 			
@@ -1306,25 +1447,81 @@ gtk_source_view_expose (GtkWidget      *widget,
 			redraw_rect.width = visible_rect.width;
 			redraw_rect.height = visible_rect.height;
 
-			gtk_paint_vline (widget->style, 
-					 event->window, 
-					 GTK_WIDGET_STATE (widget), 
-					 &redraw_rect, 
-					 widget,
-					 "margin", 
-					 redraw_rect.y, 
-					 redraw_rect.y + redraw_rect.height, 
-					 view->priv->cached_margin_width -
-					 visible_rect.x + redraw_rect.x +
-					 gtk_text_view_get_left_margin (text_view));
+			cr = gdk_cairo_create (gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_TEXT));
+			
+			/* Set a clip region for the expose event. */
+			cairo_rectangle (cr, event->area.x, event->area.y,
+					 event->area.width, event->area.height);
+			cairo_clip (cr);
+
+			/* Offset with 0.5 is needed for a sharp line. */
+			x = view->priv->cached_margin_width -
+				visible_rect.x + redraw_rect.x + 0.5 +
+				gtk_text_view_get_left_margin (text_view);
+
+			/* Default line width is 2.0 which is too wide. */
+			cairo_set_line_width (cr, 1.0);
+
+			cairo_move_to (cr, x, redraw_rect.y);
+			cairo_line_to (cr, x, redraw_rect.y + redraw_rect.height);
+			
+			gtk_widget_style_get (widget,
+					      "right-margin-line-alpha", &alpha,
+					      "right-margin-line-color", &line_color,
+					      "right-margin-overlay-toggle", &toggle,
+					      NULL);
+
+			if (!line_color)
+				line_color = gdk_color_copy (&widget->style->text[GTK_STATE_NORMAL]);
+			
+			cairo_set_source_rgba (cr,
+					       line_color->red / 65535.,
+					       line_color->green / 65535.,
+					       line_color->blue / 65535.,
+					       alpha / 255.);
+			gdk_color_free (line_color);
+			cairo_stroke (cr);
+			
+			/* g_strstrip doesn't allocate a new string. */
+			toggle = g_strstrip (toggle);
+			
+			/* Only draw the overlay when the theme explicitly permits it. */
+			if (g_ascii_strcasecmp ("TRUE", toggle) == 0 ||
+			    strcmp ("1", toggle) == 0)
+			{
+				GdkColor *overlay_color;
+				
+				gtk_widget_style_get (widget,
+						      "right-margin-overlay-alpha", &alpha,
+						      "right-margin-overlay-color", &overlay_color,
+						      NULL);
+
+				if (!overlay_color)
+					overlay_color = gdk_color_copy (&widget->style->text[GTK_STATE_NORMAL]);
+
+				/* Draw the rectangle next to the line (x+.5). */
+				cairo_rectangle (cr,
+						 x + .5,
+						 redraw_rect.y,
+						 redraw_rect.width - x - .5,
+						 redraw_rect.y + redraw_rect.height);
+				cairo_set_source_rgba (cr,
+						       overlay_color->red / 65535.,
+						       overlay_color->green / 65535.,
+						       overlay_color->blue / 65535.,
+						       alpha / 255.);
+				gdk_color_free (overlay_color);
+				cairo_fill (cr);
+			}
+
+			g_free (toggle);
+			cairo_destroy (cr);
+			
+			PROFILE ({
+				g_timer_stop (timer);
+				g_message ("Time to draw the margin: %g (sec * 1000)", g_timer_elapsed (timer, NULL) * 1000);
+			});
 		}
-
-		
-		if (GTK_WIDGET_CLASS (parent_class)->expose_event)
-			event_handled = 
-				(* GTK_WIDGET_CLASS (parent_class)->expose_event)
-				(widget, event);
-
 	}
 	
 	return event_handled;	
@@ -1763,8 +1960,187 @@ compute_indentation (GtkSourceView *view,
 	return gtk_text_iter_get_slice (&start, &end);
 }
 
+static void
+indent_lines (GtkSourceView *view, GtkTextIter *start, GtkTextIter *end)
+{
+	GtkTextBuffer *buf;
+	gint start_line, end_line;
+	gint i;
+	gchar *tab_buffer = NULL;
+
+	buf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
+
+	start_line = gtk_text_iter_get_line (start);
+	end_line = gtk_text_iter_get_line (end);
+
+	if ((gtk_text_iter_get_visible_line_offset (end) == 0) &&
+	    (end_line > start_line))
+	{
+		end_line--;
+	}
+
+	if (gtk_source_view_get_insert_spaces_instead_of_tabs (view))
+	{
+		gint tabs_size;
+
+		tabs_size = view->priv->tabs_width;
+		tab_buffer = g_strnfill (tabs_size, ' ');
+	}
+	else
+	{
+		tab_buffer = g_strdup ("\t");
+	}
+
+	gtk_text_buffer_begin_user_action (buf);
+
+	for (i = start_line; i <= end_line; i++)
+	{
+		GtkTextIter iter;
+
+		gtk_text_buffer_get_iter_at_line (buf, &iter, i);
+
+		/* don't add indentation on empty lines */
+		if (gtk_text_iter_ends_line (&iter))
+			continue;
+
+		gtk_text_buffer_insert (buf, &iter, tab_buffer, -1);
+	}
+
+	gtk_text_buffer_end_user_action (buf);
+
+	g_free (tab_buffer);
+
+	gtk_text_view_scroll_mark_onscreen (GTK_TEXT_VIEW (view),
+					    gtk_text_buffer_get_insert (buf));
+}
+
+static void
+unindent_lines (GtkSourceView *view, GtkTextIter *start, GtkTextIter *end)
+{
+	GtkTextBuffer *buf;
+	gint start_line, end_line;
+	gint i;
+
+	buf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
+
+	start_line = gtk_text_iter_get_line (start);
+	end_line = gtk_text_iter_get_line (end);
+
+	if ((gtk_text_iter_get_visible_line_offset (end) == 0) &&
+	    (end_line > start_line))
+	{
+		end_line--;
+	}
+
+	gtk_text_buffer_begin_user_action (buf);
+
+	for (i = start_line; i <= end_line; i++)
+	{
+		GtkTextIter iter, iter2;
+
+		gtk_text_buffer_get_iter_at_line (buf, &iter, i);
+
+		if (gtk_text_iter_get_char (&iter) == '\t')
+		{
+			iter2 = iter;
+			gtk_text_iter_forward_char (&iter2);
+			gtk_text_buffer_delete (buf, &iter, &iter2);
+		}
+		else if (gtk_text_iter_get_char (&iter) == ' ')
+		{
+			gint spaces = 0;
+
+			iter2 = iter;
+
+			while (!gtk_text_iter_ends_line (&iter2))
+			{
+				if (gtk_text_iter_get_char (&iter2) == ' ')
+					spaces++;
+				else
+					break;
+
+				gtk_text_iter_forward_char (&iter2);
+			}
+
+			if (spaces > 0)
+			{
+				gint tabs = 0;
+				gint tabs_size = view->priv->tabs_width;
+
+				tabs = spaces / tabs_size;
+				spaces = spaces - (tabs * tabs_size);
+
+				if (spaces == 0)
+					spaces = tabs_size;
+
+				iter2 = iter;
+
+				gtk_text_iter_forward_chars (&iter2, spaces);
+				gtk_text_buffer_delete (buf, &iter, &iter2);
+			}
+		}
+	}
+
+	gtk_text_buffer_end_user_action (buf);
+
+	gtk_text_view_scroll_mark_onscreen (GTK_TEXT_VIEW (view),
+					    gtk_text_buffer_get_insert (buf));
+}
+
+static void
+insert_tab_or_spaces (GtkSourceView *view, GtkTextIter *start, GtkTextIter *end)
+{
+	GtkTextBuffer *buf;
+	gchar *tab_buf;
+
+	if (view->priv->insert_spaces)
+	{
+		gint tabs_size;
+		GtkTextIter iter;
+		gint cur_pos;
+		gint tab_pos;
+		gint num_of_equivalent_spaces;
+
+		tabs_size = view->priv->tabs_width; 
+
+		iter = *start;
+
+		cur_pos = gtk_text_iter_get_line_offset (start);
+		tab_pos = cur_pos;
+
+		/* If there are tabs in the line take them into account */
+		while (tab_pos > 0)
+		{
+			gunichar c;
+
+			gtk_text_iter_backward_char (&iter);
+			c = gtk_text_iter_get_char (&iter);
+			if (c == '\t')
+				break;
+			tab_pos--;
+		}
+
+		num_of_equivalent_spaces = tabs_size - (cur_pos - tab_pos) % tabs_size; 
+
+		tab_buf = g_strnfill (num_of_equivalent_spaces, ' ');
+	}
+	else
+	{
+		tab_buf = g_strdup ("\t");
+	}
+
+	buf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
+
+	gtk_text_buffer_begin_user_action (buf);
+	gtk_text_buffer_delete (buf, start, end);
+	gtk_text_buffer_insert (buf, start, tab_buf, -1);
+	gtk_text_buffer_end_user_action (buf);
+
+	g_free (tab_buf);
+}
+
 static gboolean
-key_press_event (GtkWidget *widget, GdkEventKey *event)
+gtk_source_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 {
 	GtkSourceView *view;
 	GtkTextBuffer *buf;
@@ -1780,7 +2156,9 @@ key_press_event (GtkWidget *widget, GdkEventKey *event)
 	mark = gtk_text_buffer_get_insert (buf);
 	gtk_text_buffer_get_iter_at_mark (buf, &cur, mark);
 
-	if ((key == GDK_Return) && !(event->state & GDK_SHIFT_MASK) && view->priv->auto_indent)
+	if ((key == GDK_Return || key == GDK_KP_Enter) &&
+	    !(event->state & GDK_SHIFT_MASK) &&
+	    view->priv->auto_indent)
 	{
 		/* Auto-indent means that when you press ENTER at the end of a
 		 * line, the new line is automatically indented at the same
@@ -1816,34 +2194,141 @@ key_press_event (GtkWidget *widget, GdkEventKey *event)
 		}
 	}
 
-	if ((key == GDK_Tab) && view->priv->insert_spaces)
+	/* if tab or shift+tab:
+	 * with shift+tab key is GDK_ISO_Left_Tab (depends on X?)
+	 */
+	if ((key == GDK_Tab || key == GDK_KP_Tab || key == GDK_ISO_Left_Tab) &&
+	    (event->state == 0 || event->state == GDK_SHIFT_MASK))
 	{
-		gint cur_pos;
-		gint num_of_equivalent_spaces;
-		gint tabs_size;
-		gchar *spaces;
+		GtkTextIter s, e;
+		gboolean has_selection;
 
-		tabs_size = view->priv->tabs_width; 
-		
-		cur_pos = gtk_text_iter_get_line_offset (&cur);
+		has_selection = gtk_text_buffer_get_selection_bounds (buf, &s, &e);
 
-		num_of_equivalent_spaces = tabs_size - (cur_pos % tabs_size); 
+		if (view->priv->indent_on_tab)
+		{
+			/* shift+tab: always unindent */
+			if (event->state & GDK_SHIFT_MASK)
+			{
+				unindent_lines (view, &s, &e);
+				return TRUE;
+			}
 
-		spaces = g_strnfill (num_of_equivalent_spaces, ' ');
-		
-		gtk_text_buffer_begin_user_action (buf);
-		gtk_text_buffer_insert (buf,  &cur, spaces, num_of_equivalent_spaces);
-		gtk_text_buffer_end_user_action (buf);
+			/* tab: if we have a selection which spans one whole line
+			 * or more, we mass indent, if the selection spans less then
+			 * the full line just replace the text with \t
+			 */
+			if (has_selection &&
+			    ((gtk_text_iter_starts_line (&s) && gtk_text_iter_ends_line (&e)) ||
+			     (gtk_text_iter_get_line (&s) != gtk_text_iter_get_line (&e))))
+			{
+				indent_lines (view, &s, &e);
+				return TRUE;
+			}
+		}
+ 
+		insert_tab_or_spaces (view, &s, &e);
+ 		return TRUE;
+	}
 
-		gtk_text_view_scroll_mark_onscreen (GTK_TEXT_VIEW (widget),
-						    gtk_text_buffer_get_insert (buf));
-		
-		g_free (spaces);
+	return (* GTK_WIDGET_CLASS (parent_class)->key_press_event) (widget, event);
+}
 
+static void
+extend_selection_to_line (GtkTextBuffer *buf, GtkTextIter *line_start)
+{
+	GtkTextIter start;
+	GtkTextIter end;
+	GtkTextIter line_end;
+
+	gtk_text_buffer_get_selection_bounds (buf, &start, &end);
+
+	line_end = *line_start;
+	gtk_text_iter_forward_to_line_end (&line_end);
+
+	if (gtk_text_iter_compare (&start, line_start) < 0)
+	{
+		gtk_text_buffer_select_range (buf, &start, &line_end);
+	}
+	else if (gtk_text_iter_compare (&end, &line_end) < 0)
+	{
+		/* if the selection is in this line, extend
+		 * the selection to the whole line */
+		gtk_text_buffer_select_range (buf, &line_end, line_start);
+	}
+	else
+	{
+		gtk_text_buffer_select_range (buf, &end, line_start);
+	}
+}
+
+static void
+select_line (GtkTextBuffer *buf, GtkTextIter *line_start)
+{
+	GtkTextIter iter;
+
+	iter = *line_start;
+
+	if (!gtk_text_iter_ends_line (&iter))
+		gtk_text_iter_forward_to_line_end (&iter);
+
+	/* Select the line, put the cursor at the end of the line */
+	gtk_text_buffer_select_range (buf, &iter, line_start);
+}
+
+static gboolean
+gtk_source_view_button_press_event (GtkWidget *widget, GdkEventButton *event)
+{
+	GtkSourceView *view;
+	GtkTextBuffer *buf;
+	int y_buf;
+	GtkTextIter line_start;
+
+	view = GTK_SOURCE_VIEW (widget);
+	buf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widget));
+
+	if (view->priv->show_line_numbers &&
+	    (event->window == gtk_text_view_get_window (GTK_TEXT_VIEW (view),
+						       GTK_TEXT_WINDOW_LEFT)))
+	{
+		gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (view),
+						       GTK_TEXT_WINDOW_LEFT,
+						       event->x, event->y,
+						       NULL, &y_buf);
+
+		gtk_text_view_get_line_at_y (GTK_TEXT_VIEW (view),
+					     &line_start,
+					     y_buf,
+					     NULL);
+	
+		if (event->type == GDK_BUTTON_PRESS && (event->button == 1))
+		{
+			if ((event->state & GDK_CONTROL_MASK) != 0)
+			{
+				/* Single click + Ctrl -> select the line */
+				select_line (buf, &line_start);
+			}
+			else if ((event->state & GDK_SHIFT_MASK) != 0)
+			{
+				/* Single click + Shift -> extended current
+				   selection to include the clicked line */
+				extend_selection_to_line (buf, &line_start);
+			}
+			else
+			{
+				gtk_text_buffer_place_cursor (buf, &line_start);
+			}
+		}
+		else if (event->type == GDK_2BUTTON_PRESS && (event->button == 1))
+		{
+			select_line (buf, &line_start);
+		}
+
+		/* consume the event also on right click etc */
 		return TRUE;
 	}
 
-	return	(* GTK_WIDGET_CLASS (parent_class)->key_press_event) (widget, event);
+	return GTK_WIDGET_CLASS (parent_class)->button_press_event (widget, event);
 }
 
 /**
@@ -1926,6 +2411,51 @@ gtk_source_view_set_insert_spaces_instead_of_tabs (GtkSourceView *view, gboolean
 	g_object_notify (G_OBJECT (view), "insert_spaces_instead_of_tabs");
 }
 
+/**
+ * gtk_source_view_get_indent_on_tab:
+ * @view: a #GtkSourceView.
+ *
+ * Returns whether when the tab key is pressed the current selection
+ * should get indented instead of replaced with the \t character.
+ *
+ * Return value: %TRUE if the selection is indented when tab is pressed.
+ *
+ * Since: 1.8
+ **/
+gboolean
+gtk_source_view_get_indent_on_tab (GtkSourceView *view)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_VIEW (view), FALSE);
+
+	return view->priv->indent_on_tab;
+}
+
+/**
+ * gtk_source_view_set_indent_on_tab:
+ * @view: a #GtkSourceView.
+ * @enable: whether to indent a block when tab is pressed.
+ *
+ * If %TRUE, when the tab key is pressed and there is a selection, the
+ * selected text is indented of one level instead of being replaced with
+ * the \t characters. Shift+Tab unindents the selection.
+ *
+ * Since: 1.8
+ **/
+void
+gtk_source_view_set_indent_on_tab (GtkSourceView *view, gboolean enable)
+{
+	g_return_if_fail (GTK_IS_SOURCE_VIEW (view));
+
+	enable = (enable != FALSE);
+
+	if (view->priv->indent_on_tab == enable)
+		return;
+
+	view->priv->indent_on_tab = enable;
+
+	g_object_notify (G_OBJECT (view), "indent_on_tab");
+}
+
 static void 
 view_dnd_drop (GtkTextView *view, 
 	       GdkDragContext *context,
